
"""
revisionist.util: utils for subversion revisionist tools
(c) 2007 Ben Smith-Mannschott <benpsm@gmail.com> 

License
  GNU Lesser General Public License.
  http://www.gnu.org/licenses/lgpl.html
"""

import re

pat_blank = re.compile(r'^\s*$')
pat_lead_space = re.compile(r'^(\s*)(.*)$')

def crop_text_block(text):
    """
    Strip the leading whitespace form the lines of the string text.
    The first line with leading white space determines the maximal
    amount of whitespace stripped from each line.
    """
    lines = text.splitlines()

    # get the index of the first non-blank line
    i = 0
    while i < len(lines) and pat_blank.match(lines[i]): i += 1
    first = i

    if first == len(lines):
        # it's all blank. just return the empty string
        return ""

    # get the index of the last non-blank 
    i = len(lines) - 1
    while i >= 0 and pat_blank.match(lines[i]): i -= 1
    last = i

    assert first <= last

    # cut out leading trailing blank lines
    lines = lines[first:last+1]

    # split each line into its leading whitespace portion
    # and its rest
    split_lines = [ pat_lead_space.match(x).groups()
                    for x in lines ]

    # find the amount of common leading whitespace, which is the
    # length of the shortest leading whitespace segment, ignoring
    # those with length 0.

    cut_length = 0
    for space, rest in split_lines:
        n = len(space)
        if 0 < n < cut_length or cut_length == 0:
            cut_length = n

    # remove cut_length characters from the leading whitespace of each
    # line. Lines with no leading whitespaces are effectively left
    # unaltered by this.

    lines = [ space[cut_length:] + rest
              for space, rest in split_lines ]

    # the result is the incoming lines joined by \n.  We'll want to
    # if the original text had a terminating \n, we'll want to make
    # sure that comes along by stuffing an empty string at the end
    # lines before doing the join.

    if text[-1] == '\n':
        lines.append('')
    
    result = "\n".join(lines)

    return result


def test_strip_leading_space():
    sls = strip_leading_space
    assert sls("") == ""
    assert sls("  ") == ""
    assert sls("first\n  second\n  third") == "first\nsecond\nthird"


def curry(function, *curry_args, **curry_kwargs):
    def _curried(*call_args, **call_kwargs):
        args = curry_args + call_args
        kwargs = curry_kwargs.copy()
        kwargs.update(call_kwargs)
        return function(*args, **kwargs)
    return _curried


class odict(dict):
    """
    An extension of dict which remembers the insertion order of its
    keys.  
    """
    def __init__(self, *args, **kwargs):
        """
        An odict may be initialized with content from the following
        sources:

        1. a single odict
        2. a single dict (key order undefined)
        3. a single list of (key, value) tuples
        4. an arbitrary number of (key, value) tuples
        5. an arbitrary number of keyword arguments (key order
           undefined)
        """
        dict.__init__(self)
        self.ordered_keys = []
        if len(args) == 1 and 'keys' in dir(args[0]):
            contents = args[0]
            for k in contents.keys():
               self[k] = contents[k]
        elif len(args) == 1 and 'append' in dir(args[0]):
            for k, v in args[0]:
                self[k] = v
        elif len(args) > 0:
            for k, v in args:
                self[k] = v
        for k in kwargs:
            self[k] = kwargs[k]
        
    def __setitem__(self, key, value):
        if key not in self:
            self.ordered_keys.append(key)
        dict.__setitem__(self, key, value)
        
    def __getitem__(self, key):
        return dict.__getitem__(self, key)
    
    def __delitem__(self, key):
        dict.__delitem__(self, key)
        self.ordered_keys.remove(key)
        
    def keys(self):
        return list(self.ordered_keys)
    
    def copy(self):
        cp = odict(self)
        return cp

    def iteritems(self):
        for k in self.ordered_keys:
            yield k, self[k]

    def __iter__(self):
        return iter(self.ordered_keys)

    def iterkeys(self):
        return iter(self.ordered_keys)

    def itervalues(self):
        for k in self.ordered_keys:
            yield self[k]

    def __str__(self):
        out = ["{"]
        for k in self.ordered_keys:
            if len(out) > 1:
                out.append(', ')
            out.append('%s: %s' % (k, self[k]))
        out.append("}")
        return "".join(out)

    def __repr__(self):
        return "odict(%r)" % ( [item for item in self.iteritems()],)

                   
def test_odict():
    o = odict()
    assert len(o) == 0
    o = odict([ (5,1), (4,2), (3,3), (2,4), (1,5) ])
    assert o.keys() == [5, 4, 3, 2, 1]
    o = odict((5,1),(4,2),(3,3),(2,4),(1,5))
    assert o.keys() == [5, 4, 3, 2, 1]
    d = dict([(5,1),(4,2),(3,3),(2,4),(1,5)])
    assert d.keys() != [5, 4, 3, 2, 1]
    o2 = odict(o)
    assert o.keys() == o2.keys()
    assert [x for x in o.iteritems()] == [(5,1),(4,2),(3,3),(2,4),(1,5)]
    assert [x for x in o.iterkeys()] == [5, 4, 3, 2, 1]
    assert [x for x in o.itervalues()] == [1, 2, 3, 4, 5]
    
            
    
    



"""
revisionist.editors: edit subversion dumpfiles emitted by parser
(c) 2007 Ben Smith-Mannschott <benpsm@gmail.com> 

License
  GNU Lesser General Public License.
  http://www.gnu.org/licenses/lgpl.html
"""

import sys
from util import crop_text_block as msg
from parser import BeginDumpfile, EndDumpfile, \
     BeginRevision, EndRevisionHeader, EndRevisionNodes, \
     BeginNode, EndNode, UserProperties, TextContent, BlankLine


def edit_properties(events, edit):
    """
    Modifies parse events.  Consumes a stream of parse events,
    invoking the function 'edit' on BeginRevsion, BeginNode and
    UserProperties events before yielding the events to its caller.

    The events passed to edit are all dictionaries. Edit may modify
    their content as it sees fit.

    Changes to UserProperties will automatically cause recomputation
    of Content-length of Prop-content-length of the owning Node or
    Revision.
    """
    evt = events.next()
    while type(evt) != EndDumpfile:
        if type(evt) in (BeginRevision, BeginNode) :
            edit(evt) # Edit dump properties of Node or Revision
            
            # we'll need to postpone emitting this event and any that
            # follow until the end of the node or end of the revision
            # header so that we can recompute Prop and Content lengths
            # if UserProperties have been changed.

            evt_hold = [evt]
            prop_evt = None
            evt = events.next()
            while type(evt) not in (EndRevisionHeader, EndNode):
                if type(evt) == UserProperties:
                    assert prop_evt == None
                    prop_evt = evt
                evt_hold.append(evt)
                evt = events.next()

            assert type(evt) in (EndRevisionHeader, EndNode)
            if prop_evt:
                assert type(evt_hold[0]) in (BeginRevision, BeginNode)
                # edit user properties of node or Revision
                edit(prop_evt)
                # recompute Prop-content-length and Content-length
                dump_props = evt_hold[0]
                prop_len = len(str(prop_evt))
                dump_props["Prop-content-length"] = prop_len
                text_len = int(dump_props.get("Text-content-length", 0))
                dump_props["Content-length"] = prop_len + text_len
            for held_evt in evt_hold:
                yield held_evt
            yield evt
        else:
            yield evt
        evt = events.next()
    assert type(evt) == EndDumpfile
    yield evt

def echo_properties(events, property_names):
    """
    Print selected properties to stderr as they pass through.
    
    Passes parse events through unchanged.  Any property, be it a user
    property or a dump property with a name in the list events will be
    printed.
    """
    for evt in events:
        if type(evt) in [ UserProperties, BeginNode, BeginRevision ]:
            for name in property_names:
                if name in evt.keys():
                    print >>sys.stderr, name, evt[name]
        yield evt

def consume_events(events):
    """
    The universal sink. Think of it as /dev/null for parse events.
    """
    for x in events:
        pass
    
def show_progress(events):
    """
    Print out some nead lines of periods to stderr while events pass
    through.
    """
    period = 1000
    line_width = 78
    
    next_nl = period*line_width
    next_dot = period
    n = 0
    for evt in events:
        yield evt
        n += 1
        if n == next_dot:
            sys.stderr.write(".")
            next_dot += period
            if n > next_nl:
                sys.stderr.write("\n")
                next_nl += period*line_width
    else:
        sys.stderr.write("\n")





#!/usr/bin/env python

"""
Test script for revisionist module.
(c) 2007 Ben Smith-Mannschott <benpsm@gmail.com> 

License
  GNU Lesser General Public License.
  http://www.gnu.org/licenses/lgpl.html
"""

from util import crop_text_block as msg
import parser
import writer
import editors
import os

dumpfiles = ["short.dump2",  "short.dump3"]

def round_trip_test(dumpFilePath):
    inFilePath = dumpFilePath
    outFilePath = dumpFilePath + ".out"
    inFile = file(inFilePath, "rb")
    outFile = file(outFilePath, "wb")
    events = parser.pull(inFile)
    writer.write_events_to_dumpfile(events, outFile)
    inFile.close()
    outFile.close()

    original = file(inFilePath, "rb")
    copy = file(outFilePath, "rb")

    original_bytes = original.read()
    copy_bytes = copy.read()

    original.close()
    copy.close()

    assert original_bytes == copy_bytes

    os.unlink(outFilePath)

def run_tests():
    for filePath in dumpfiles:
        round_trip_test(filePath)
    print "ok"

def main():
    run_tests()

if __name__ == "__main__":
    main()

    




#!/usr/bin/env python

import sys
from revisionist import *

def parse_options():
    "Parse command line options. See also print_usage."
    args = sys.argv[1:]
    args.append(None)
    if args[0] in ["-h", "--help", None]:
        print_usage()
        return None, None
    verbose = False
    propsubs = []
    while args[0] in ["--property", "-p", "--verbose", "-v"]:
        if args[0] in ["--property", "-p"]:
            del args[0]
            propname = args[0]; del args[0]
            replacements = []
            while args[0] in ["--replace", "-r",
                              "--normalize-line-breaks", "-n"]:
                if args[0] in ["--replace", "-r"]:
                    del args[0]
                    replacements.append((args[0], args[1]))
                    del args[0]
                    del args[0]
                else:
                    del args[0]
                    replacements.append(('\r\n', '\n'))
            propsubs.append((propname, replacements))
        elif args[0] in ["--verbose", "-v"]:
            del args[0]
            verbose = True
    if len(args) != 1 or args[0] != None:
        print_usage()
        return []
    else:
        return propsubs, verbose


def main():
    propsubs, verbose = parse_options()
    if propsubs == None:
        return 1

    def edit(props):
        for propname, replacements in propsubs:
            # covers two cases: propname not present; propname has
            # value none which indicates a V3 style propdiff deletion.
            val = props.get(propname, None)
            if val:
                for old, new in replacements:
                    val = val.replace(old, new)
                props[propname] = val
                
    propnames = [propname for propname, x in propsubs]
    events = parser.pull(sys.stdin)
    if verbose:
        events = editors.echo_properties(events, propnames)
    events = editors.edit_properties(events, edit)
    if verbose:
        events = editors.echo_properties(events, propnames)
    writer.write_events_to_dumpfile(events, sys.stdout)

def print_usage():
    print >>sys.stderr, \
"""
 %s OPTIONS < dumpfile.in > dumpfile.out

 Legal option combinations are described by this BNF:

 OPTIONS         = HelpOpt | PropertyClause* 
 HelpOpt         = -h | --help
 PropertyClause  = PropertyOpt EditClause*
 PropertyOpt     = -p | --property
 EditClause      = NormalizeOpt | ReplaceClause
 NormalizeOpt    = -n | --normalize-line-breaks
 ReplaceClause   = ReplaceOpt OldText NewText
 ReplaceOpt      = -r | --replace
 OldText         = text
 NewText         = text
 
 e.g.

 %s --property svn:externals \\
   --replace svn://old.com/repos/ svn://new.com/repos/ -n

 1. Replace every occurance of the string 'svn://old.com/repos/' with
    'svn://new.com/repos' in every svn:externals property in the
    dumpfile.
 2. Normalize the line breaks in every svn:externals property.
""" % (sys.argv[0], sys.argv[0])


if __name__ == "__main__":
    sys.exit(main())




"""
revisionist.parser: a parser for Subversion dumpfiles (vers 2 & 3).
(c) 2007 Ben Smith-Mannschott <benpsm@gmail.com> 

License
  GNU Lesser General Public License.
  http://www.gnu.org/licenses/lgpl.html
"""

import re
import sys
from md5 import md5
from util import crop_text_block as msg
from util import curry, odict

# ------------------------------------------------------------------------
# This module's primary entry point(s)
# ------------------------------------------------------------------------

def pull(fileLike):
    """
    Parse the SVN Dumpfile in the open file fileLike.

    This is a generator. It yields a series of parse events described
    by the following BNF-like notation::

      BeginDumpfile
      BlankLine* 
        (
        BeginRevision
        BlankLine?
        UserProperties?
        BlankLine*
        EndRevisionHeader
          (
          BeginNode
          BlankLine?
          UserProperties?
            (
            TextContent
            BlankLine
            )?
          BlankLine*
          )*
        EndRevisionNodes
        BlankLine*
        )*
      EndDumpFile
    """
    return Parser().parse(fileLike)

# ------------------------------------------------------------------------
# Parse Events
# ------------------------------------------------------------------------

class BeginDumpfile( object ):
    """
    Parse Event. Signalling the start of a Dumpfile.

    version
        Version of the dumpfile. This parse is written for dumpfile
        formats 2 and 3, though the parser doesn't check for this.

    uuid
        The UUID of the repository.  This can be None.
    """
    def __init__(self, version, uuid=None):
        self.version = version
        self.uuid = uuid
    def __str__(self):
        out = ["SVN-fs-dump-format-version: %d\n\n" % (self.version,)]
        if (self.uuid):
            out.append("UUID: %s\n" % (self.uuid,))
        return "".join(out)
    def __repr__(self):
        return "BeginDumpfile(%r, %r)" % (self.version, self.uuid)

        
class EndDumpfile( object ):
    """
    Parse Event. Signals the end of the Dumpfile.  No further events
    will follow.
    """
    def __str__(self):
        return ""
    def __repr__(self):
        return "EndDumpfile()"


class BeginRevision( odict ):
    """
    Parse Event. Signals the start of a Revision.

    This is a dictionary. It's items are the dumpfile properties of
    the the revision, for example Revision-number.

    keys() preserves the order present in the original dump file.
    """
    def __str__(self):
        """
        Returns the contained dumpfile properties exactly as they
        should be written to the dumpfile.
        """
        return "".join(["%s: %s\n" % (k, self[k]) for k in self.keys()])
    def __repr__(self):
        return "BeginRevision(%s)" % odict.__repr__(self)


class EndRevisionHeader(object):
    """
    Parse Event. Signals that the Revision's header has been parsed.
    """
    def __str__(self):
        return ""
    def __repr__(self):
        return "EndRevisionHeader()"


class EndRevisionNodes(object):
    """
    Parse Event. Signals the end of the Revision.  Only EndDumpfile or
    BeginRevision may follow.
    """
    def __str__(self):
        return ""
    def __repr__(self):
        return "EndRevisionNodes()"


class BeginNode( odict ):
    """
    Parse Event. Signals the start of a Node.

    This is a dictionary. It's items are the dumpfile properties of
    the node, for ecample: Node-path.

    keys() preserves the order of declaration.
    """
    def __str__(self):
        """
        Returns the contained dumpfile properties exactly as they
        should be written to the dumpfile.
        """
        return "".join(["%s: %s\n" % (k, self[k]) for k in self.keys()])
    def __repr__(self):
        return "BeginNode(%s)" % odict.__repr__(self)


class EndNode( object ):
    """
    Parse Event. Signals the end of a Node.
    """
    def __str__(self):
        return ""
    def __repr__(self):
        return "EndNode()"


class UserProperties( odict ):
    """
    Parse Event. Signals user properties, i.e. Node or Revision
    properties in the SVN sense.

    This is a dictionary.  It's items are the user parsed user
    properties, e.g. 'svn:externals', 'svn:log', etc.

    keys() preserves declaration order.

    This may occur between BeginNode and EndNode or between
    BeginRevision and EndRevisionHeader.
    """
    def __str__(self):
        out = []
        for k in self.keys():
            v = self[k]
            if v != None:
                out.append("K %d\n%s\nV %d\n%s\n" % (len(k), k, len(v), v))
            else:
                out.append("D %d\n%s\n" % (len(k), k))
        else:
            out.append("PROPS-END\n")
        return "".join(out)
    def __repr__(self):
        return "UserProperties(%s)" % odict.__repr__(self)


class TextContent( str ):
    """
    Parse Event. Signals text content of a node. (Revisions never have
    text content.)

    It behaves like a string.  It not only signals the content, it
    *is* the content.
    """
    def __repr__(self):
        return "TextContent(%s)" % str.__repr__(self)


class BlankLine(object):
    """
    Parse Event. Signals a blank line between two other Events.
    """
    def __init__(self):
        pass
    def __str__(self):
        return '\n'
    def __repr__(self):
        return "BlankLine()"


# ------------------------------------------------------------------------
# The Pull Parser
# ------------------------------------------------------------------------

pat_dump_property = re.compile(r"^([-A-Za-z0-9_]+):[ ](.*)$")
        
class Parser(object):
    """
    A parser for Subversion dump file format versions 2 and 3.

    This is a *pull* parser, meaning it presents itself as a generator
    offering a series of parse events.

    The parse events and the sequence in which you can expect to
    receive them is not described here. See the function parse() of
    this module.

    This parser is intentionally *very* unforgiving.  It will stop
    noisily with an AssertionError if anything about its input doesn't
    meet with its expectations.

    It's better to fail obviously than to silently corrupt your data,
    after all.  (Incidentally, being so explicit about my assumptions
    and understandings of the dump file format also made it much
    easier to develop a *correct* parser.)
    """
    
    def __init__(self):
        pass
    
    def parse(self, fileLike):
        """
        Generate parse events from the bytes provided by fileLike.
        """
        try:
            self.reader = None
            self.reader = Reader(fileLike)
            self.reader.next()
            for evt in self.parseDumpfile(): yield evt
        except:
            print >>sys.stderr, str(self.reader)
            raise
        assert self.reader.eof, msg("""
            Stopped parsing before end of input
            %s""" % (parser.reader,))

    def parseDumpfile(self):
        """
        A Dumpfile consists of a Version, an (optional?) UUID, and
        zero or more Revisions.  Extra blank lines may occur.  Blank
        lines before and after the version declaration are not
        reported.
        """
        self.version = None
        self.skipBlankLines()
        version = int(self.parseDumpProperty("SVN-fs-dump-format-version"))
        assert 2 <= version <= 3, \
               "Only dump format versions 2 and 3 are supported"
        self.version = version
        self.skipBlankLines()
        if self.matchDumpProperty("UUID"):
            uuid = self.parseDumpProperty("UUID")
        else:
            uuid = None

        yield BeginDumpfile(version, uuid)

        for evt in self.parseBlankLines(): yield evt
            
        while self.matchRevision():
            for evt in self.parseRevision(): yield evt
            for evt in self.parseBlankLines(): yield evt

        yield EndDumpfile()

    def matchDumpProperty(self, name=None):
        if name:
            prefix = name + ": "
            result = self.reader.cur.startswith(prefix)
            return result
        else:
            return pat_dump_property.match(self.reader.cur)
    
    def parseDumpProperty(self, name=None, store=None):
        """
        A dump property consists of a name and a value on a single
        line, separated by a colon.

        name
           If specified: (1) the parsed property must have this
           name. (2) we return the value of the property as a string.

           If None, then we return a tuple (*name*, *value*), both
           strings.

        store
           If specified, it must be a dictionary.  We'll update it as
           store[name] = value before we return.

        returns
          *value* or ( *name*, *value* )
        """
        m = pat_dump_property.match(self.reader.cur)

        assert m, "Expecting a dump property, but found \n%s" % (self.reader,)
        assert name == None or name == m.group(1), msg("""
            Expected property %s, but found %s
            """ % (name, m.group(1)))
        
        value = m.group(2)
        if store != None:
            store[m.group(1)] = value
        self.reader.next()
        if name == None:
            return m.group(1), value
        else:
            return value

    def skipBlankLines(self):
        """
        Skip blank lines in input without reporting them as parse events.
        """
        while self.reader.cur == "\n":
            self.reader.next()

    def matchBlankLine(self):
        return self.reader.cur == "\n"
    
    def parseBlankLine(self):
        """
        Parse blank line in input, reporting it as a BlankLine()
        event.
        """
        assert self.reader.cur == "\n", "Expected blank line"
        self.reader.next()
        return BlankLine()

    def parseBlankLines(self):
        while self.reader.cur == "\n":
            self.reader.next()
            yield BlankLine()


    def getBytes(self, n):
        """
        Returns the next n bytes of input, starting with and including
        the current line.

        We consume one extra byte of input without reporting it. It is
        always a line feed character which terminates the value.
        """
        buf, buflen = [], 0
        while n >= buflen and not self.reader.eof:
            buflen += len(self.reader.cur)
            buf.append(self.reader.cur)
            self.reader.next()
        assert buflen == n+1 and buf[-1][-1] == '\n', \
            "Didn't find expected newline terminator."
        buf[-1] = buf[-1][:-1] # strip newline terminator
        result = "".join(buf)
        assert len(result) == n
        return result

    def matchRevision(self):
        return self.matchDumpProperty("Revision-number")
    
    def parseRevision(self):
        """
        Parse a revision, which may contain UserProperties and will
        not include TextContent::

          Revision = 
            BeginRevision
            BlankLine?
            UserProperties?
            BlankLine*
            EndRevisionHeader
              (
              BeginNode
              BlankLine?
              UserProperties?
                (
                TextContent
                BlankLine
                )?
              BlankLine*
              )*
            EndRevisionNodes
        """
        dump_props = odict()
        rev = int(self.parseDumpProperty("Revision-number", dump_props))
        plen = int(self.parseDumpProperty("Prop-content-length", dump_props))
        clen = int(self.parseDumpProperty("Content-length", dump_props))
        assert clen - plen == 0, "A revision never has text content."

        yield BeginRevision(dump_props)

        if self.matchBlankLine():
            yield self.parseBlankLine()
        if self.matchUserProperties():
            yield self.parseUserProperties(plen)
        for evt in self.parseBlankLines():
            yield evt

        yield EndRevisionHeader()

        while self.matchNode():
            for evt in self.parseNode():
                yield evt

        yield EndRevisionNodes()

    def matchUserProperties(self):
        return (self.matchUserPropertyKey()
                or self.matchUserPropertyDelete())
    
    def parseUserProperties(self, plen, prop_delta=False):
        start = self.reader.start
        properties = UserProperties()
        while not self.matchUserPropertyEnd(): 
            assert not self.reader.eof
            if self.matchUserPropertyKey():
                key = self.parseUserPropertyKey()
                value = self.parseUserPropertyValue()
                properties[key] = value
            elif self.matchUserPropertyDelete():
                assert prop_delta, msg("""
                    Property deletion (operation 'D') is only allowed
                    when the Prop-delta dump property is true for the
                    containing node or revision.
                    """)
                key = self.parseUserPropertyDelete()
                properties[key] = None
        else:
            self.parseUserPropertyEnd()
        stop = self.reader.start
        assert plen == stop - start
        return properties

    def matchUserPropertyEnd(self):
        return self.reader.cur == "PROPS-END\n"

    def parseUserPropertyEnd(self):
        assert self.matchUserPropertyEnd(), msg("""
            Expected PROPS-END, but found\n%s
            """ % (self.reader,))
        self.reader.next()
    
    def matchUserPropertyKey(self):
        return self.reader.cur.startswith("K ")

    def parseUserPropertyKey(self):
        assert self.matchUserPropertyKey()
        key = self.getPropertyEntryContent()
        return key

    def matchUserPropertyValue(self):
        return self.reader.cur.startswith("V ")

    def parseUserPropertyValue(self):
        assert self.matchUserPropertyValue()
        value = self.getPropertyEntryContent()
        return value

    def matchUserPropertyDelete(self):
        return self.reader.cur.startswith("D ")

    def parseUserPropertyDelete(self):
        assert self.matchUserPropertyDelete()
        key = self.getPropertyEntryContent()
        return key

    def getPropertyEntryContent(self):
        """
        Parses property entry of the form:

           chararcter space n:integer newline bytes[n] newline

        and returns the bytes[n].
        """
        assert (len(self.reader.cur) and
                self.reader.cur[0:2] in ['K ', 'V ', 'D ']), msg("""
                Expected a pair of property entry lines, where the
                first has the form (K|V|D) <number>. Found this:
                %s""" % (self.reader,))
        n = int(self.reader.cur[2:])
        self.reader.next()
        result = self.getBytes(n)
        return result

    def matchNode(self):
        return self.matchDumpProperty("Node-path")
                                 
    def parseNode(self):
        """
        Parse a single node, which may contain both UserProperties and
        TextContent.
        
          Node = 
              BeginNode
              BlankLine?
              UserProperties?
                (
                TextContent
                BlankLine
                )?
        """
        chunk_pos = self.reader.start
        dump_props = BeginNode()
        node_path = self.parseDumpProperty("Node-path", dump_props)
        if self.matchDumpProperty("Node-kind"):
            node_kind = self.parseDumpProperty("Node-kind", dump_props)
        else:
            node_kind = None
        node_action = self.parseDumpProperty("Node-action", dump_props)
        
        # A number of optional properties follow. We know what they
        # can be, but their order isn't fixed.
        #
        # Content-length is optional, contrary to the available
        # documentation, which claims that it is always present for
        # compatability with generic rfc822 parsers.
        #
        # e.g. a dir node with action add and no properties has no
        # lengths at all.

        tlen, plen, clen = None, None, None
        while self.matchDumpProperty():
            name, value = self.parseDumpProperty(store=dump_props)
            if name == "Text-content-length":
                tlen = int(value)
            elif name == "Prop-content-length":
                plen = int(value)
            elif name == "Content-length":
                clen = int(value)

        prop_delta = dump_props.get("Prop-delta") == "true"
        if prop_delta:
            assert self.version > 2, msg("""
                Property deltas should not occur in this dumpfile.
                Its format is too old to support them.
                """)

        text_delta = dump_props.get("Text-delta") == "true"
        if text_delta:
            assert self.version > 2, msg("""
                Text deltas should not occur in this dumpfile.
                Its format is too old to support them.
                """)

        if clen == None:
            clen = 0
        if plen == None:
            plen = 0
        if tlen != None:
            assert tlen == clen - plen
        else:
            tlen = clen - plen

        yield dump_props

        if plen > 0 or tlen > 0:
            yield self.parseBlankLine()

        if plen > 0:
            yield self.parseUserProperties(plen, prop_delta)

        if tlen > 0:
            text = self.getBytes(tlen)
            assert len(text) == tlen

            # We can only verify the checksum when text_deltas are not
            # in use.  When text_deltas are being used, the checksum
            # refers to the *result* of applying the deltas and we
            # have no idea how nor desire to do that here.
            if not text_delta:
                expected_chksum = dump_props.get("Text-content-md5")
                if expected_chksum:
                    computed_chksum = md5(text).hexdigest()
                    assert expected_chksum == computed_chksum, msg("""
                           MD5 mismatch.
                           expected: %s,
                           computed: %s.
                           """ % (expected_chksum, computed_chksum))

            yield TextContent(text)
            # TextContent is always terminated by an 'extra' newline,
            # which getBytes consumes for us, but does not return.
            yield BlankLine() 

        for evt in self.parseBlankLines():
            yield evt

        yield EndNode()


class Reader(object):
    """
    Reads a fileLike object one line at a time, while remembering the
    most recently read line in 'cur'.  This allows us a look-ahead of
    one line, which is all we need.

    It keeps tack of quite a lot of state (linenr, start, stop, ...)
    for ease of debugging.

    fileLike
        The source of text lines

    cur
        The most recently read line

    start
        The byte offset, within fileLike of the first byte of
        cur. (i.e. cur[0])

    stop
        The byte offset, within fileLike, of the the next character
        following the last character of cur.  (This makes start
        .. stop a half-open range.)

    eof
        True fileLike's supply of lines has been exhausted.

    linenr
        The number (1-based) of the line of text in cur.
    """
    def __init__(self, fileLike):
        """
        Initialize a new LineReader.

        - fileLike must be a file open for reading.
        - fileLike will be closed automatically when all lines have
          been consumed.
        """
        self.fileLike = fileLike
        self.cur = None
        self.start = 0
        self.stop = 0
        self.eof = False
        self.linenr = 0

    def next(self):
        """
        Read the next line of input, returning it and storing it in
        cur.  Other properties are updated accordingly.  The first
        call to next(), following the last line will set eof to True.
        """
        try:
            self.cur = self.fileLike.next()
            self.linenr += 1
            self.start, self.stop = self.stop, self.stop+len(self.cur)
        except StopIteration:
            self.cur = ""
            self.eof = True
            self.start, self.stop = self.stop, self.stop
            self.fileLike.close()
        return self.cur

    def close(self):
        """
        Close the underlying fileLike
        """
        self.fileLike.close()

    def __str__(self):
        """
        The str(Reader) is intended for debugging.
        """
        return msg("""
            LineReader
            cur[%3d] = %s
            start    = %d
            stop     = %d
            linenr   = %d
            eof      = %s
            """ % (len(self.cur), self.cur[:72], 
                   self.start, self.stop, self.linenr, self.eof))





#!/usr/bin/env python

from distutils.core import setup
setup(
    name='revisionist',
    version='1.0.0',
    description='Library to parse and modify Subversion dumpfiles.',
    author="Ben Smith-Mannschott",
    author_email="benpsm@gmail.com",
    packages=["revisionist"],
    scripts=['revisionist-fixprops.py'],
    package_data={'revisionist': ['*.dump2', '*.dump3']}
    )
    
    




"""
revisionist.write: write Subversion dumpfiles (vers 2 & 3)
(c) 2007 Ben Smith-Mannschott <benpsm@gmail.com> 

License
  GNU Lesser General Public License.
  http://www.gnu.org/licenses/lgpl.html
"""

import sys
from md5 import md5
from util import crop_text_block as msg
from parser import BeginDumpfile, EndDumpfile, \
     BeginRevision, EndRevisionHeader, EndRevisionNodes, \
     BeginNode, EndNode, UserProperties, TextContent, BlankLine


def write_events_to_dumpfile(events, dstFile):
    """
    Consume a series of parse events while writing them dstFile as a
    SVN Dumpfile.

    All events will return exactly the bytes that need to be written
    to the dumpfile when they are asked for their string
    representation.
    """
    version = None
    text_content_md5 = None
    text_content_length = None

    try:
        for evt in events:
            
            # This is all sanity checking, to make sure we don't
            # silently produce an invalid dump file.
            
            if type(evt) == BeginDumpfile:
                # remember the version of the dump file
                version = evt.version
                assert 2 <= version <= 3, msg(
                    """Only dumpfile format versions 2 and 3 are supported.""")

            if version == 2:
                # Version 2 doesn't support Text and Property deltas,
                # so make sure we're not using any of those.
                if type(evt) == UserProperties:
                    assert None not in evt.values(), msg(
                        """Property deleting requires Prop-deltas: true
                           and dump file format version 3 or higher.""")
                if type(evt) in (BeginRevision, BeginNode):
                    assert evt.get("Prop-delta") != "true", msg(
                        """Dump file format must be at least version 3 to
                        support Prop-deltas.""")
                if type(evt) == BeginNode:
                    assert evt.get("Text-delta") != "true", msg(
                        """Text deltas should not occur in this dumpfile.
                           Its format is too old to support them.
                           version = %s""" % (version,))

            if type(evt) == BeginNode:
                # remember checksum and lengths
                if evt.get("Text-delta") == "true":
                    # don't bother to remember the checksum.  we don't
                    # know how to interpret text deltas.
                    text_content_md5 = None
                else:
                    text_content_md5 = evt.get("Text-content-md5")
                text_content_length = int(evt.get("Text-content-length", 0))
                prop_content_length = int(evt.get("Prop-content-length", 0))
                
            elif type(evt) == EndNode:
                # forget checksum and size
                text_content_md5 = None
                text_content_length = None
                prop_content_length = None

            elif type(evt) == BeginRevision:
                prop_content_length = int(evt.get("Prop-content-length", 0))

            elif type(evt) == EndRevisionHeader:
                prop_content_length = None
                
            elif type(evt) == TextContent:
                # validate against text_content_length
                assert text_content_length == len(evt), msg(
                    """Text length mismatched.
                       Text-content-length: %d
                       Actual length:       %d"""
                    % ( text_content_length, len(evt) ))
                # validate the text content against the checksum, if
                # present.
                if text_content_md5:
                    h = md5(evt).hexdigest()
                    assert h == text_content_md5, msg(
                           """MD5 mismatch. The Text-content-md5 claimed by
                              the parent node does not match the computed md5.
                              expected: %s
                              computed: %s"""
                           % (text_content_md5, h))

            elif type(evt) == UserProperties:
                assert prop_content_length == len(str(evt)), msg(
                    """Property length mismatched.
                       Prop-content-length: %d
                       Actual length:       %d"""
                    % (prop_content_length, len(str(evt))))
                    
            # this is the actual write to the file. compact, isn't it?
                    
            dstFile.write(str(evt))
            
    finally:
        dstFile.close()
